from .boolean import Boolean
from .color import Color
from .girder_folder import GirderFolder
from .girder_item import GirderItem
from .integer import Integer
from .number import Number
from .number_choice import NumberChoice
from .number_multichoice import NumberMultichoice
from .number_vector import NumberVector
from .range import Range
from .string import String
from .string_choice import StringChoice
from .string_multichoice import StringMultichoice
from .string_vector import StringVector

__all__ = (
    Boolean,
    Color,
    GirderFolder,
    GirderItem,
    Integer,
    Number,
    NumberChoice,
    NumberMultichoice,
    NumberVector,
    Range,
    String,
    StringChoice,
    StringMultichoice,
    StringVector
)



from inspect import getdoc
try:
    from inspect import signature
except ImportError:  # pragma: nocover
    from funcsigs import signature


class MissingDescriptionException(Exception):
    """Raised when a function is missing description decorators."""


class MissingInputException(Exception):
    """Raised when a required input is missing."""


def get_description_attribute(func):
    """Get the private description attribute from a function."""
    func = getattr(func, 'run', func)
    description = getattr(func, '_girder_description', None)
    if description is None:
        raise MissingDescriptionException('Function is missing description decorators')
    return description


def argument(name, data_type, *args, **kwargs):
    """Describe an argument to a function as a function decorator.

    Additional arguments are passed to the type class constructor.

    :param str name: The parameter name from the function declaration
    :param type: A type class derived from ``girder_worker_utils.types.Base``
    """
    if not isinstance(name, str):
        raise TypeError('Expected argument name to be a string')

    if callable(data_type):
        data_type = data_type(name, *args, **kwargs)

    def argument_wrapper(func):
        func._girder_description = getattr(func, '_girder_description', {})
        args = func._girder_description.setdefault('arguments', [])
        sig = signature(func)

        if name not in sig.parameters:
            raise ValueError('Invalid argument name "%s"' % name)

        data_type.set_parameter(sig.parameters[name], signature=sig)
        args.insert(0, data_type)

        def call_item_task(inputs, outputs=None):
            args, kwargs = parse_inputs(func, inputs)
            return func(*args, **kwargs)

        def describe():
            return describe_function(func)

        func.call_item_task = call_item_task
        func.describe = describe

        return func

    return argument_wrapper


def describe_function(func):
    """Return a json description from a decorated function."""
    description = get_description_attribute(func)

    inputs = [arg.describe() for arg in description.get('arguments', [])]
    spec = {
        'name': description.get('name', func.__name__),
        'inputs': inputs,
        'mode': 'girder_worker'
    }
    desc = description.get('description', getdoc(func))
    if desc:
        spec['description'] = desc

    return spec


def get_input_data(arg, input_binding):
    """Parse an input binding from a function argument description.

    :param arg: An instantiated type description
    :param input_binding: An input binding object
    :returns: The parameter value
    """
    mode = input_binding.get('mode', 'inline')
    if mode == 'inline' and 'data' in input_binding:
        value = arg.deserialize(input_binding['data'])
    elif mode == 'girder':
        value = input_binding.get('id')
    else:
        raise ValueError('Unhandled input mode')

    arg.validate(value)
    return value


def parse_inputs(func, inputs):
    """Parse an object of input bindings from item_tasks.

    :param func: The task function
    :param dict inputs: The input task bindings object
    :returns: args and kwargs objects to call the function with
    """
    description = get_description_attribute(func)
    arguments = description.get('arguments', [])
    args = []
    kwargs = {}
    for arg in arguments:
        desc = arg.describe()
        input_id = desc['id']
        name = desc['name']
        if input_id not in inputs and not arg.has_default():
            raise MissingInputException('Required input "%s" not provided' % name)
        if input_id in inputs:
            kwargs[name] = get_input_data(arg, inputs[input_id])
    return args, kwargs



from girder_worker.app import app
from girder_worker.utils import types
from girder_worker.utils.decorators import argument


def not_a_task():
    pass


@argument('n', types.Integer)
def function_task(n):
    return n


@app.task
@argument('n', types.Integer)
def celery_task(n):
    return n



from importlib import import_module
import celery
from girder_worker.utils import decorators

from stevedore import extension


#: Defines the namespace used for plugin entrypoints
NAMESPACE = 'girder_worker_plugins'


def _handle_entrypoint_errors(mgr, entrypoint, exc):
    raise exc


def get_extension_manager(app=None):
    """Get an extension manager for the plugin namespace."""
    if app is None:
        app = celery.current_app

    return extension.ExtensionManager(
        namespace=NAMESPACE,
        invoke_on_load=True,
        invoke_args=(app,),
        on_load_failure_callback=_handle_entrypoint_errors
    )


def get_plugin_task_modules(app=None):
    """Return task modules defined by plugins."""
    includes = []
    for ext in get_extension_manager(app=app):
        includes.extend(ext.obj.task_imports())
    return includes


def import_all_includes():
    """Import all task modules for their side-effects."""
    for module in get_plugin_task_modules():
        import_module(module)


def get_extensions(app=None):
    """Get a list of installed extensions."""
    return [ext.name for ext in get_extension_manager(app)]


def get_module_tasks(module_name):
    """Get all tasks defined in a python module.

    :param str module_name: The importable module name
    """
    module = import_module(module_name)
    tasks = {}

    if module is None:
        return tasks

    for name, func in vars(module).items():
        full_name = '%s.%s' % (module_name, name)
        if not callable(func):
            # filter out objects that are not callable
            continue

        try:
            decorators.get_description_attribute(func)
            tasks[full_name] = func
        except decorators.MissingDescriptionException:
            pass
    return tasks


def get_extension_tasks(extension, app=None, celery_only=False):
    """Get the tasks defined by a girder_worker extension.

    :param str extension: The extension name
    :param app: The celery app instance
    :param bool celery_only: If true, only return celery tasks
    """
    manager = get_extension_manager(app)
    imports = manager[extension].obj.task_imports()
    tasks = {}
    for module_name in imports:
        tasks.update(get_module_tasks(module_name))

    if celery_only:  # filter celery tasks
        if app is None:
            from .app import app
        tasks = {
            key: tasks[key] for key in tasks if key in app.tasks
        }

    return tasks


def discover_tasks(app):
    app.conf.update({
        'CELERY_INCLUDE': get_plugin_task_modules()
    })



from .number import Number
from .vector import Vector


class NumberVector(Vector):
    """Define a parameter accepting a list of numbers.

    >>> @argument('value', types.NumberVector, min=10, max=100, step=10)
    ... def func(value=(10, 11)):
    ...     pass
    """

    paramType = 'number-vector'
    elementClass = Number



import os
import re
import shutil
import setuptools

from setuptools.command.install import install


def prerelease_local_scheme(version):
    """Return local scheme version unless building on master in CircleCI.

    This function returns the local scheme version number
    (e.g. 0.0.0.dev<N>+g<HASH>) unless building on CircleCI for a
    pre-release in which case it ignores the hash and produces a
    PEP440 compliant pre-release version number (e.g. 0.0.0.dev<N>).
    """
    from setuptools_scm.version import get_local_node_and_date

    # this regex allows us to publish pypi packages from master, our LTS maintenance branches, and
    # our next major version integration branches
    pattern = r'master|[0-9]+\.x-maintenance|v[0-9]+-integration'

    if re.match(pattern, os.getenv('CIRCLE_BRANCH', '')):
        return ''
    else:
        return get_local_node_and_date(version)


class CustomInstall(install):
    """
    Override the default install to add some custom install-time behavior.
    Namely, we create the local config file.
    """

    def run(self, *args, **kwargs):
        install.run(self, *args, **kwargs)

        distcfg = os.path.join('girder_worker', 'worker.dist.cfg')
        localcfg = os.path.join('girder_worker', 'worker.local.cfg')
        if not os.path.isfile(localcfg):
            print('Creating worker.local.cfg')
            shutil.copyfile(distcfg, localcfg)


with open('requirements.in') as f:
    install_reqs = f.readlines()

setuptools.setup(
    name='girder-worker',
    use_scm_version={'root': '..', 'local_scheme': prerelease_local_scheme},
    setup_requires=['setuptools_scm', 'setuptools-git'],
    description='Batch execution engine built on celery.',
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    url='https://github.com/girder/girder_worker',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    packages=setuptools.find_packages(
        exclude=('tests.*', 'tests')
    ),
    include_package_data=True,
    cmdclass={
        'install': CustomInstall
    },
    install_requires=install_reqs,
    python_requires='>=3.8',
    zip_safe=False,
    entry_points={
        'girder_worker_plugins': [
            'docker = girder_worker.docker:DockerPlugin [docker]'
        ],
        'girder_worker._test_plugins.valid_plugins': [
            'plugin1 = girder_worker._test_plugins.plugins:TestPlugin1',
            'plugin2 = girder_worker._test_plugins.plugins:TestPlugin2'
        ],
        'girder_worker._test_plugins.invalid_plugins': [
            'exception1 = girder_worker._test_plugins.plugins:TestPluginException1',  # noqa
            'exception2 = girder_worker._test_plugins.plugins:TestPluginException2',  # noqa
            'import = girder_worker._test_plugins.plugins:TestPluginInvalidModule',  # noqa
            'invalid = girder_worker._test_plugins.plugins:NotAValidClass'
        ],
    }
)



import os
import shutil
import tempfile

from ..girder_io import GirderClientTransform


class GirderFileIdAllowDirect(GirderClientTransform):
    """
    This transform either uses direct path to access a file, if possible and
    allowed, or downloads a Girder File to the local machine and passes its
    local path into the function.  The direct path is only used if the
    GW_DIRECT_PATHS environment variable is set.

    WARNING: if a direct path is used, the task MUST NOT modify the file.  It
    is the resposibility of the user of this transform to ensure tasks treat
    files as read-only.

    To use this transform from Girder, it should be called via something like:
    ```
    try:
        local_path = File().getLocalFilePath(file)
    except FilePathException:
        local_path = None
    input_path = GirderFileIdAllowDirect(
        str(file['_id']), file['name'], local_path)
    ```

    :param _id: The ID of the file to download.
    :type _id: str
    :param name: The name of the file.  If the file must be downloaded, the
        extension is preserved.
    :type name: str
    :param local_path: If specified and the path exists and is reachable by
        after the transform, the file is accessed directly.
    :type local_path: str
    """

    def __init__(self, _id, name='', local_path=None, **kwargs):
        super().__init__(**kwargs)
        self.file_id = _id
        self.file_name = name
        self.local_file_path = local_path

    def _repr_model_(self):
        if self.local_file_path:
            return '{}({!r}, {!r}, {!r})'.format(
                self.__class__.__name__, self.file_id, self.file_name, self.local_file_path)
        return f'{self.__class__.__name__}({self.file_id!r}, {self.file_name!r})'

    def _allowDirectPath(self):
        """
        Check if the worker environment permits direct paths.  This just checks
        if the environment variable GW_DIRECT_PATHS is set to a non-empty
        value.

        :returns: True if direct paths are allowed.
        """
        return bool(os.environ.get('GW_DIRECT_PATHS'))

    def transform(self):
        # Don't download if self.local_file_path is set and direct paths are
        # allowed.
        if (self.local_file_path and self._allowDirectPath()
                and os.path.isfile(self.local_file_path)):
            self.temp_dir_path = None
            self.file_path = self.local_file_path
        else:
            self.temp_dir_path = tempfile.mkdtemp()
            self.file_path = os.path.join(self.temp_dir_path, '{}{}'.format(
                self.file_id, os.path.splitext(self.file_name)[1]))
            self.gc.downloadFile(self.file_id, self.file_path)
        return self.file_path

    def cleanup(self):
        if self.temp_dir_path:
            shutil.rmtree(self.temp_dir_path, ignore_errors=True)
